/* ------------------------------------------------------
	typist

    This program will print out exercise lines to the
    display, and then check for correct typing.
---------------------------------------------------------
	Original: D. Jason Penney (penneyj@slc.com)
	Tuned for ECIP NeXT: Takeshi Ogihara
		(ogihara@rd.ecip.osaka-u.ac.jp)
	Ver.1.0   1992-07-14
	Ver.1.1   1992-08-05	ESC  ->  Ctrl-F
	Ver.1.2   1994-11-10	for non-NeXT WS
	Ver.1.21  1994-12-07	do_para() bug
	Ver.1.3   1994-12-09	REPEAT & BELL
------------------------------------------------------ */

/* NeXT  ... This macro should be predefined on NeXTSTEP. */

#include <stdio.h>
#include <ctype.h>
#include <signal.h>

#if defined(ultrix)
# include <cursesX.h>
#else
# include <curses.h>
#endif

#if defined(VMS) || defined(ultrix)
# define cbreak() crmode()
# define nocbreak() nocrmode()
#endif

#include <errno.h>

#include "typist.h"

#define ASCII_ESC	27
#define ASCII_GS	29
#define CTRL_F		6
#define ASCII_BEL	7
#define ASCII_FF	12
#define ASCII_BS	8
#define ASCII_RUBOUT	127

#if defined(VMS)
# define ESCAPER ASCII_GS
#else
# define ESCAPER CTRL_F		/* was ASCII_ESC */
#endif

#define MAX_LINES	1500
static char *cached_lines[MAX_LINES];

#ifdef JPN
# define Fputs(s)  fputs(s, stderr)

#if defined(__alpha) || defined(ultrix)
# define jaddstr addstr
#else
static void jaddstr(s)
    char *s;
{
    refresh();
    fputs(s, stderr);
    fflush(stderr);
}
#endif
#endif

static int linesInLesson;
static int exerciseStart;
static int newStart;

static FILE *lesson_file;
static StrType current_line;
static BoolType BreakLesson, SkipLesson;

static int wait_user()
/* Get any response from user before proceeding */
{
    int ch, help=0;

    for ( ; ; ) {
	move(23, 0);
	standout();
#ifdef JPN
	jaddstr("˿ʤˤϲ򲡤Ʋ...");
#else
	addstr("Press any key to continue...");
#endif
	standend();
	refresh();
	ch = getch();
	if (ch == ASCII_BEL) {
	    move(0, 0);
	    standout();
	    addstr(loud ? "[BELL OFF]":"[BELL  ON]");
	    standend();
	    loud = !loud;
	}else if (ch == '?' && help == 0) {
	    move(0, 0);
	    standout();
	    addstr("[ *HELP* ] ^F=SKIP ^H/BS=REPEAT ^L=MENU ^G=BELL ^C=QUIT");
	    standend();
	    help = 1;
	}else
	    break;
    }
    return ch;
}

static void get_end_line()
/* Read end of line, should be smarter... */
{
    refresh();  getch();
}

static int cleanup()	/* cleanup routine, esp. SIGINT */
{
    move(23,0);
    addstr("See you again!!");
    refresh();
    endwin();
    write_rc();
    printf("\n");
    exit(1);
}

static char *get_lesson()	/* Ask user for desired lesson */
{
    static StrType response;
    int ch, n;
    register struct Lesson *p;

    if (LastLesson) {
	ch = LastLesson;
	n = (LastNum == FINISHED)? 1: (LastNum + 1);
    }else {
	ch = typeLessons[0].nam;
	n = 1;
    }

    for ( ; ; ) {
	move(0,0);
	refresh();
#ifdef JPN
	Fputs("  ʲ饳򤷤Ʋ:\n");
	for (p = typeLessons; p->item; p++) {
	    fprintf(stderr,"        %s (%c1 - %c%d)",
			p->item, p->nam, p->nam, p->num);
	    if (p->fin == 0)
		fprintf(stderr, "\n");
	    else if (p->fin >= p->num)
		fprintf(stderr, " !!!λ!!!\n");
	    else
		fprintf(stderr, " [%c%dޤǽλ]\n", p->nam, p->fin);
	}
	fprintf(stderr,
	    "\n å̾( %c%d)ϤƲ",
	    ch, n);
	if (HasHelp) {
	    Fputs("\n");
	    jaddstr(" (λ=ԡإ=?) ------> ");
	}else {
	    fflush(stderr);
	    jaddstr("(λ=): ");
	}
#else
	Fputs("  Several lessons are available:\n");
	for (p = typeLessons; p->item; p++) {
	    fprintf(stderr,"        %s (%c1 - %c%d)\n",
			p->item, p->nam, p->nam, p->num);
	    if (p->fin == 0)
		fprintf(stderr, "\n");
	    else if (p->fin >= p->num)
		fprintf(stderr, " !!!Finish!!!\n");
	    else
		fprintf(stderr, " [%c%d was done]\n",
			p->nam, p->fin);
	}
	fprintf(stderr,
	    "\n Type the desired lesson name(e.g. %c%d).\n",
	    ch, n);
	jaddstr(" (Quit=RETURN) ------------> ");
#endif

	refresh();
	echo();
	nocbreak();

# ifndef NeXT
	(void)getstr(response);
# else
	(void)gets(response);
# endif

	noecho();
	cbreak();
	if (strlen(response) == 0)
	    cleanup();	/* EXIT */
#ifdef JPN
	if (HasHelp && response[0] == '?')
	    return "H1";	/* HELP FILE */
#endif
	break;
    }
    return response;
}

static BoolType get_lesson_line()
/* Read a line of input from user, in canonical mode */
{
    if (feof(lesson_file) || ferror(lesson_file))
	return FALSE;
    if (fgets(current_line, STR_SIZE, lesson_file) == NULL)
	current_line[0] = '\0';
    else {
	register int x = strlen(current_line) - 1;
	if (current_line[x] == '\n')
	    current_line[x] = '\0';
    }
    return TRUE;
}

static BoolType find_lesson(name)
/* locate given lesson, leave lesson_file open so next line begins */
    char *name;
{
    StrType fullName;
    StrType target;
    StrType response;
    BoolType done = FALSE;

    if (!CheckLessonName(name)) {
#ifdef JPN
	fprintf(stderr,
		"\n \"%s\" ȤåϤޤ.\n", name);
#else
	fprintf(stderr, "\n %s: No such Lesson.\n", name);
#endif
	(void) wait_user();
	return FALSE;
    }
    /* gf: Added LessonDir */
    sprintf(fullName, "%s/%c.typ", LessonDir, name[0]);
    sprintf(target, "*%c%d*", toupper(name[0]), atoi(&name[1]));
    lesson_file = fopen(fullName, "r");
    if (lesson_file == NULL) {
	standout();
	sprintf(response,
		"OPEN ERROR: %s (errno=%d)\n", fullName, errno);
	addstr(response);
	standend();
	(void) wait_user();
	return FALSE;
    }
    while (!done) {
	if (!get_lesson_line())
	    break;
	if (strncmp(current_line, target, 5) == 0) {
	    done = TRUE;
	    break;
	}
    }
    if (!done) {
#ifdef JPN
	sprintf(response, "å%s ѤǤޤ.\n", &name[1]);
	jaddstr(response);
#else
	sprintf(response, "Lesson %s is not available.\n", &name[1]);
	standout();
	addstr(response);
	standend();
#endif
	(void) wait_user();
	return FALSE;
    }
    linesInLesson = 0;
    done = FALSE;
    while (!done) {
	if (!get_lesson_line()) {
	    done = TRUE;
	    break;
	}
	if (current_line[0] == '*') {
	    done = TRUE;
	    break;
	}
	if (linesInLesson == MAX_LINES) {
	}
	if (cached_lines[linesInLesson] != NULL) {
	    free(cached_lines[linesInLesson]);
	    cached_lines[linesInLesson] = NULL;
	}
	cached_lines[linesInLesson] = (char *)malloc(strlen(current_line) + 1);
	if (cached_lines[linesInLesson] == NULL) {
	    endwin();
	    fprintf(stderr, "SYSTEM ERROR: malloc() failure\n");
	    write_rc();
	    exit(1);
	}
	strcpy(cached_lines[linesInLesson], current_line);
	linesInLesson ++;
    }
    return TRUE;
}

static void get_exercise(startLine, kind, newStart)
/* determine length and kind of exercise */
    int startLine;
    char *kind;
    int *newStart;
{
    char *thisLine;
    int lineLength;

    *newStart = startLine;
    *kind = 'I'; /* default... */
    /* Determine end of exercise */
    while (TRUE) {
	if (*newStart >= linesInLesson)
	    break; /* done */

	thisLine = cached_lines[*newStart];
	lineLength = strlen(thisLine);
	if (thisLine[lineLength - 2] == '\\') { /* expected terminator */
	    *kind = thisLine[lineLength - 1];
	    thisLine[lineLength - 2] = '\0'; /* remove */
	    (*newStart) ++;
	    break;
	}
	if (thisLine[lineLength - 1] == '\\') /* compatibility with BASIC */
	    thisLine[lineLength - 1] = '\0'; /* also remove */
	(*newStart) ++;
    }
}

static void displaySpeed(totalChars, elapsed, errs)
    /* show how fast you were */
    int totalChars;
    int elapsed;
    int errs;
{
    double testTime, words, speed, adjustedSpeed;
    StrType message;

    if (elapsed <= 0) elapsed = 1;
    testTime = (double)elapsed / (double)60.0; /* time in minutes */
    words = (double)totalChars / 5.0;
    speed = words / testTime;
    words -= errs;
    adjustedSpeed = words / testTime;

    move(19, 19);
    standout();
#ifdef JPN
    sprintf(message, "ϥԡ       = %.2f (/ʬ)", speed);
    jaddstr(message);
    move(20, 19);
    sprintf(message, "Ϥ = %.2f (/ʬ)", adjustedSpeed);
    jaddstr(message);
    move(21, 19);
    sprintf(message, "              (Ψ %.2f)",
	(double)100.0 * errs / totalChars);
    jaddstr(message);
#else
    sprintf(message, "raw speed      = %.2f wpm", speed);
    addstr(message);
    move(20, 19);
    sprintf(message, "adjusted speed = %.2f wpm", adjustedSpeed);
    addstr(message);
    move(21, 19);
    sprintf(message, "            with %.2f%s errors",
	(double)100.0 * errs / totalChars, "%");
    addstr(message);
#endif
    standend();
}

static void do_drill()	/* \D, a drill */
{
    register int ch, i,j;
    int tries, oncemore;
    int errors;
    int totalChars;
    char *thisLine;
    int thisLength;
    long startTime, endTime;

    for (tries = 0; ; tries++) { /* a try */
	totalChars = 0;
	errors = 0;

	/* display drill pattern */
	for (i = exerciseStart; i < newStart; i ++) {
	    move(3 + (i - exerciseStart) * 2, 0);
	    addstr(cached_lines[i]);
	    totalChars += strlen(cached_lines[i]) + 1; /* +1 for NULL */
	}
	refresh();

	/* run the drill */
	for (i = exerciseStart; i < newStart; i ++) {
	    thisLine = cached_lines[i];
	    thisLength = strlen(thisLine);

	    move(4 + (i - exerciseStart) * 2, 0);
	    for (j = 0; j < thisLength; j ++) {
		refresh();
		ch = getch();
		if (i == exerciseStart && j == 0)
		    startTime = time(NULL);
		    /* timer doesn't run till first stroke */
		if (ch < ' ') {
		    if (ch == ESCAPER) {
			clear();
			goto ALL_DONE;
		    }
		    if (ch == ASCII_FF) {
			BreakLesson = TRUE;
			clear();
			goto ALL_DONE;
		    }
		}
		if (ch != thisLine[j]) {
		    addch('X');
		    if (loud) sndbeep();
		    errors ++;
		} else
		    addch(ch);
	    }
	    get_end_line();
	}
	endTime = time(NULL);

	displaySpeed(totalChars, endTime - startTime, errors);
	ch = wait_user();
	clear();
	if (ch == ESCAPER)
	    goto ALL_DONE;
	if (ch == ASCII_FF) {
	    BreakLesson = TRUE;
	    goto ALL_DONE;
	}
	oncemore = (ch == ASCII_BS || ch == ASCII_RUBOUT);

/* Exit condition:
 *	tries = 0  =>  errors = 0
 *	tries = 1  =>  errors = 1
 *	tries = 2  =>  errors = 2
 *	tries >= 3  =>  exit
 */
	if (!oncemore && (tries >= 3 || errors <= tries))
	    break;

	move(0, 0);
#ifdef JPN
	jaddstr("⤦...");
#else
	addstr("Try again...");
#endif
    }
    return;

    ALL_DONE:
	SkipLesson = TRUE;
}

static void do_para()	/* \P, practice paragraph */
{
    register int ch, i, j;
    char *thisLine;
    int thisLength;
    int totalChars = 0;
    int errors = 0;
    int startTime, endTime;
    int oncemore = 0;
    
    do {
	if (oncemore) {
	    clear();
	    move(0, 0);
#ifdef JPN
	    jaddstr("⤦...");
#else
	    addstr("Try again...");
#endif
	}
	/* print out practice text */
	for (i = exerciseStart; i < newStart; i ++) {
	    move(3 + (i - exerciseStart), 0);
	    addstr(cached_lines[i]);
	    totalChars += strlen(cached_lines[i]);
	}
    
	move(3, 0);
	refresh();
	for (i = exerciseStart; i < newStart; i ++) {
	    move(3 + (i - exerciseStart), 0);
	    thisLine = cached_lines[i];
	    thisLength = strlen(thisLine);
	    for (j = 0; j < thisLength; j ++) {
		refresh();
		ch = getch();
		if (i == exerciseStart && j == 0)
		    startTime = time(NULL);
		    /* timer doesn't run till first stroke */
		if (ch < ' ') {
		    if (ch == ESCAPER) {
			goto ALL_DONE;
		    }
		    if (ch == ASCII_FF) {
			BreakLesson = TRUE;
			goto ALL_DONE;
		    }
		}
		/* Allow backspace */
		if (j != 0 && (ch == ASCII_BS || ch == ASCII_RUBOUT)) {
		    j --;
		    move(3 + (i - exerciseStart), j);
		    addch(thisLine[j]);
		    move(3 + (i - exerciseStart), j);
		    j --;
		    continue; /* but notice that errors accumulate */
		}
		if (ch != thisLine[j]) {
		    move(3 + (i - exerciseStart), j);
		    standout();
		    addch(thisLine[j]);
		    standend();
		    if (loud) sndbeep();
		    errors ++;
		} else
		    move(3 + (i - exerciseStart), j + 1);
	    }
	    get_end_line();
	}
	endTime = time(NULL);
    
	displaySpeed(totalChars, endTime - startTime, errors);
	if ((ch = wait_user()) == ASCII_FF) {
	    BreakLesson = TRUE;
	    goto ALL_DONE;
	}
	oncemore = (ch == ASCII_BS || ch == ASCII_RUBOUT);
    }while (oncemore);
    clear();
    return;

ALL_DONE:
    SkipLesson = TRUE;
    clear();
}

static void do_type()	/* \T, type out a lesson */
{
    int i;

    clear();
    for (i = exerciseStart; i < newStart; i ++) {
	move(1 + (i - exerciseStart), 0);
#ifdef JPN
	jaddstr(cached_lines[i]);
#else
	addstr(cached_lines[i]);
#endif
    }
    if (wait_user() == ASCII_FF)
	BreakLesson = TRUE;
    clear();
}

static void do_instruction()	/* \I, instruction line */
{
    clear();
    move(0, 0);
#ifdef JPN
    jaddstr(cached_lines[exerciseStart]);
#else
    addstr(cached_lines[exerciseStart]);
#endif
}

static void do_clear()		/* \B, clear screen */
{
    clear();
    move(1, 0);
#ifdef JPN
    jaddstr(cached_lines[exerciseStart]);
#else
    addstr(cached_lines[exerciseStart]);
#endif
    if (wait_user() == ASCII_FF)
	BreakLesson = TRUE;
}

static void lesson_loop()	/* dance through lesson file */
{
    char kind;

    exerciseStart = 0;
    while (TRUE) {
	get_exercise(exerciseStart, &kind, &newStart);
	BreakLesson = FALSE;
	switch (kind) {
	case 'B':
	  do_clear();
	  break;

	case 'P':
	  do_para();
	  break;

	case 'T':
	  do_type();
	  break;

	case 'D':
	  do_drill();
	  break;

	case 'I':
	  do_instruction();
	  break;

	default:
	  endwin();
	  fprintf(stderr, "DATA ERROR: Unknown Exercise \
(near line %d of this lesson)\n", newStart - 1);
	  write_rc();
	  exit(1);
	  break;
	}
	if (BreakLesson)
	    break;
	exerciseStart = newStart;
	if (exerciseStart >= linesInLesson) {
	    break;
	}
    }
}

static void give_lesson(lnam)
    char *lnam;
{
    register struct Lesson *p;
    int ch, n;

    SkipLesson = FALSE;
    lesson_loop();
    if (BreakLesson || SkipLesson) return;

    ch = lnam[0];
    n = atoi(&lnam[1]);
    for (p = typeLessons; p->item; p++) {
	if (ch == p->nam) {
	    LastLesson = ch;	/* ...think H */
	    if (p->num > n) {
		if (p->fin < n) p->fin = n;
		LastNum = n;
	    }else
		LastNum = p->fin = FINISHED;
	    break;
	}
    }
}


void typist(lnam)
    char *lnam;
{
    char *lessonName;
    int i;

    initscr();
    clear();
    refresh();
    cbreak();
    noecho();
    read_rc();
    sndbeep_init(LessonDir);
    signal(SIGINT, cleanup);

    for (i = 0; i < MAX_LINES; i ++)
	cached_lines[i] = NULL;

    /* Start and format screen.  Open and read exercise file.
       Inquire for lesson to run */
    if (lnam != NULL) {
	if (find_lesson(lnam)) give_lesson(lnam);
	else sleep(1);
    }

    while (TRUE) {
	clear();
	lessonName = get_lesson();
	if (!find_lesson(lessonName))
	    continue;
	give_lesson(lessonName);
    }
}
